import bpy
import os


#UI

def draw_material_settings(context, layout, node):
    
    inputs = [socket for socket in node.inputs if socket.enabled and not socket.is_linked]
    for input in inputs:
        row = layout.row()
        if input.hide_value:
            row.alignment = "CENTER"
        input.draw(context, row, node, input.name)

#Functions
def get_previews_from_files(self, context):
    #This is pretty much just straight from the template scipt "Ui Previews Dynamic Enum"
    #so thanks blender devs :)
    """EnumProperty callback"""
    enum_items = []

    if context is None:
        return enum_items

    script_dir = os.path.dirname(os.path.abspath(__file__))
    directory = os.path.join(script_dir,"maps","rendered_maps","")

    # Get the preview collection (defined in register func).
    pcoll = preview_collections["alpha_trees_previews"]

    if directory == pcoll.my_previews_dir:
        return pcoll.my_previews

    print("Scanning directory: %s" % directory)

    if directory and os.path.exists(directory):
        # Scan the directory for files
        image_paths = []
        for fn in os.listdir(directory):
            if "diff_leaf" in fn:
                image_paths.append(fn)

        for i, name in enumerate(image_paths):
            # generates a thumbnail preview for a file.
            filepath = os.path.join(directory, name)
            icon = pcoll.get(name)
            if not icon:
                thumb = pcoll.load(name, filepath, 'IMAGE')
            else:
                thumb = pcoll[name]
            label = name[:-19]
            enum_items.append((name, label, name, thumb.icon_id, i))

    pcoll.my_previews = enum_items
    pcoll.my_previews_dir = directory
    return pcoll.my_previews


def get_diff_maps():
    script_path = os.path.dirname(os.path.abspath(__file__))
    str_dir = os.path.join(script_path, "maps", "rendered_maps", "")
    dir = os.fsencode(str_dir)
    diff_maps = []
    for file in os.listdir(dir):
        filename = os.fsdecode(file)
        if "diff_leaf" in str(filename):
            diff_maps.append(filename)
    return str_dir, diff_maps


def import_material(blend_path):
    with bpy.data.libraries.load(blend_path) as (data_from, data_to):
        data_to.materials.append("alpha_trees_material")
    return(data_to.materials[0])


def import_image(image_path):
    print("")
    diff_leaf_name = os.path.basename(image_path)

    for image in bpy.data.images:
        if image.name == diff_leaf_name:
            diff_leaf_image = image
            break
    else:
        print("Loading Image")
        diff_leaf_image = bpy.data.images.load(image_path)

    ratio = diff_leaf_image.size[1]/diff_leaf_image.size[0]
    print("Ratio: " + str(ratio))
    bpy.ops.mesh.primitive_plane_add()
    plane = bpy.context.active_object
    plane.name = diff_leaf_name[:-19]

    verts = plane.data.vertices
    verts[0].co = (-1, 0, 0)
    verts[1].co = (1, 0, 0)
    verts[2].co = (-1, 0, 2*ratio)
    verts[3].co = (1, 0, 2*ratio)

    scale = float(diff_leaf_name[-18:-14])
    print("Scale: " + str(scale))
    plane.scale = (scale, scale, scale)

    bpy.ops.object.material_slot_add()
    mat_slot = plane.material_slots[0]

    for material in bpy.data.materials:
        if material.name == "alpha_trees_material":
            print("Found Material")
            base_tree_material = material
            break
    else:
        print("Loading Material")
        script_path = os.path.dirname(os.path.abspath(__file__))
        blend_path = os.path.join(script_path, "base_scene.blend")
        base_tree_material = import_material(blend_path)

    for material in bpy.data.materials:
        if material.name == diff_leaf_name[:-19]:
            tree_material = material
            break
    else:
        tree_material = base_tree_material.copy()
        tree_material.name = diff_leaf_name[:-19]

    mat_slot.material = tree_material

    # Switch images
    diff_trunk_image = bpy.data.images.load(
        image_path.replace("leaf", "trunk"))
    nor_leaf_image = bpy.data.images.load(image_path.replace("diff", "nor"))
    nor_trunk_image = bpy.data.images.load(
        image_path.replace("diff_leaf", "nor_trunk"))
    mask_image = bpy.data.images.load(
        image_path.replace("diff_leaf.jpg", "mask.png"))

    nor_leaf_image.colorspace_settings.name = "Non-Color"
    nor_trunk_image.colorspace_settings.name = "Non-Color"

    nodes = tree_material.node_tree.nodes
    for node in nodes:
        if node.type == "TEX_IMAGE":
            if node.label == "leaf color":
                node.image = diff_leaf_image

            elif node.label == "trunk color":
                node.image = diff_trunk_image

            elif node.label == "leaf normal":
                node.image = nor_leaf_image

            elif node.label == "trunk normal":
                node.image = nor_trunk_image

            elif node.label == "mask":
                node.image = mask_image

    return plane

preview_collections = {}

def register():
    #register icons
    pcoll = bpy.utils.previews.new()
    pcoll.my_previews_dir = ""
    pcoll.my_previews = ()
    preview_collections["alpha_trees_previews"] = pcoll


def unregister():
    #unregister icons
    for pcoll in preview_collections.values():
        bpy.utils.previews.remove(pcoll)
    preview_collections.clear()