import bpy
import os
import time
import webbrowser
from bpy.props import FloatProperty, IntProperty
from . import gen_functions, imp_functions

#Generator
class ALPHATREE_OT_set_up_scene(bpy.types.Operator):
    """Set up scene for rendering"""
    bl_label = "Set up scene"
    bl_idname = "alpha_tree.set_up_scene"
    bl_description = "Set up scene for rendering"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        gen_functions.setup_scene(self, context)

        return {"FINISHED"}


class ALPHATREE_OT_set_up_tree(bpy.types.Operator):
    """Set up tree for rendering"""
    bl_label = "Set up tree"
    bl_idname = "alpha_tree.set_up_tree"
    bl_description = "Set up tree for rendering"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        if not bpy.context.selected_objects:
            self.report({"WARNING"}, "No objects selected")
            return {"CANCELLED"}

        gen_functions.scale_to_height()
        gen_functions.set_render_border(context)
        gen_functions.transparent_normal(self, context)

        return {"FINISHED"}


class ALPHATREE_OT_render_maps(bpy.types.Operator):
    """render all maps"""
    bl_label = "Render maps"
    bl_idname = "alpha_tree.render_maps"
    bl_description = "render all maps"
    bl_description = "render all maps"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cscene = scene.cycles
        layout.scale_y = 1.3
        layout.prop(cscene, "samples", text="samples (50 max.)")

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):

        if not bpy.context.selected_objects:
            self.report({"WARNING"}, "No objects selected")
            return {"CANCELLED"}

        gen_functions.render_maps(self, context)

        return {"FINISHED"}


class ALPHATREE_OT_open_folder(bpy.types.Operator):
    """open the renders folder"""
    bl_label = "Open renders folder"
    bl_idname = "alpha_tree.open_folder"
    bl_description = "Open renders folder"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        at = bpy.context.scene.alpha_trees
        if at.open_leaf_folder:
            script_path = os.path.dirname(os.path.abspath(__file__))
            script_path = os.path.join(script_path,"maps", "leaf_maps", "")
            path = bpy.path.abspath(script_path)
        else:
            path = bpy.path.abspath(at.render_filepath)
        print("Opening: " + path)
        webbrowser.open('file:///' + path)
        return {"FINISHED"}


class ALPHATREE_OT_execute_all(bpy.types.Operator):
    """Execute all operators"""
    bl_label = "Execute all"
    bl_idname = "alpha_tree.execute_all"
    bl_description = "Execute all operators"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cscene = scene.cycles
        layout.scale_y = 1.3
        layout.prop(cscene, "samples", text="samples (50 max.)")

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):

        if not bpy.context.selected_objects:
            self.report({"WARNING"}, "No objects selected")
            return {"CANCELLED"}

        start = time.time()
        at = bpy.context.scene.alpha_trees

        print("Scaling object")
        gen_functions.scale_to_height()
        print("Rendering normal image")
        gen_functions.transparent_normal(self, context)
        print("Rendering maps")
        gen_functions.render_maps(self, context)
        if at.open_renders_folder:
            print("Opening folder")
            bpy.ops.alpha_tree.open_folder()

        total_time = round(time.time() - start, 2)

        print(f"Done in {total_time} seconds")

        return {"FINISHED"}

#Importer
class ALPHATREE_OT_import_tree(bpy.types.Operator):
    """Import a tree from the available library"""
    bl_label = "Import Tree"
    bl_idname = "alpha_tree.import_tree"
    bl_description = "Import a tree from the available library"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        at = context.scene.alpha_trees
        path = os.path.dirname(os.path.abspath(__file__))
        print("Importing " + at.alpha_trees_previews[:-19])
        imp_functions.import_image(os.path.join(path,"maps","rendered_maps",at.alpha_trees_previews))

        return {"FINISHED"}


classes = (
    ALPHATREE_OT_set_up_scene,
    ALPHATREE_OT_execute_all,
    ALPHATREE_OT_open_folder,
    ALPHATREE_OT_render_maps,
    ALPHATREE_OT_set_up_tree,
    ALPHATREE_OT_import_tree,
)

preview_collections = {}

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)