import bpy
import os
from . import imp_functions
from bpy.props import BoolProperty, FloatProperty, IntProperty, StringProperty, EnumProperty


class ATGeneratorSettings(bpy.types.PropertyGroup):

    # Show UI
    # ----------------------------------------------------------

    show_info: BoolProperty(
        name="Info",
        description="Show info about how to use the addon",
        default=False
    )

    show_setup_tree_settings: BoolProperty(
        name="settings",
        description="Show setup settings",
        default=False
    )

    show_render_settings: BoolProperty(
        name="Show render settings",
        description="Show render settings",
        default=False,
    )

    show_open_folder_settings: BoolProperty(
        name="settings",
        description="Show Alpha trees operators",
        default=False,
    )

    show_exec_all_settings: BoolProperty(
        name="settings",
        description="Show Alpha trees operators",
        default=False,
    )

    # generator operator settings
    # ----------------------------------------------------------

    script_path = os.path.dirname(os.path.abspath(__file__))

    z_dimension: FloatProperty(
        name="Camera height",
        default=10,
        description="Target height of the model",
        soft_min=0,
    )

    scale_padding: FloatProperty(
        name="Scale padding",
        default=0.03,
        description="Amount of extra space around object",
        soft_min=0,
        soft_max=0.1,
    )

    border_padding: FloatProperty(
        name="Render border padding",
        default=0.03,
        description="Amount of extra space around object",
        soft_min=0,
        soft_max=0.2,
    )

    resolution: IntProperty(
        name="Max resolution",
        default=1024,
        description="Max resolution of images",
        subtype="PIXEL"
    )

    render_filepath: StringProperty(
        name="",
        default=os.path.join(script_path,"maps","rendered_maps",""),
        description="Filepath to save rendered images to",
        subtype="FILE_PATH"
    )

    diff_render: BoolProperty(
        name="render diffuse map",
        default=True,
        description="render the diffuse map",
    )

    nor_render: BoolProperty(
        name="render normal map",
        default=True,
        description="render the normal map",
    )

    mask_render: BoolProperty(
        name="render mask map",
        default=True,
        description="render the mask map",
    )

    overwrite: BoolProperty(
        name="Overwrite",
        default=True,
        description="Overwrite images if they already exist",
    )

    remove_extra_masks: BoolProperty(
        name="delete extra masks",
        default=True,
        description="delete extra masks after combining (off for debug)",
    )

    open_renders_folder: BoolProperty(
        name="open renders folder on completion",
        default=False,
        description="open the renders folder on completion",
    )

    open_leaf_folder: BoolProperty(
        name="open leaf folder",
        default=False,
        description="open the leaf folder instead",
    )

    # importer operator settings
    # ----------------------------------------------------------

    alpha_trees_previews : EnumProperty(
        items=imp_functions.get_previews_from_files,
    )

def register():
    bpy.utils.register_class(ATGeneratorSettings)
    bpy.types.Scene.alpha_trees = bpy.props.PointerProperty(
        type=ATGeneratorSettings)


def unregister():
    bpy.utils.unregister_class(ATGeneratorSettings)
    del bpy.types.Scene.alpha_trees