import bpy
from . import gen_functions, imp_functions, operators, settings

class AlphaTreesPrefs(bpy.types.AddonPreferences):
    bl_idname = __package__

    show_extra_operators: bpy.props.BoolProperty(
        name="Show extra operators",
        default = True,
    )

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "show_extra_operators")

class ALPHATREE_PT_import_panel(bpy.types.Panel):
    """Import panel"""
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_region_type = "UI"
    bl_label = "Alpa trees importer"
    bl_category = "Alpha Trees"

    def draw(self, context):
        layout = self.layout
        at = bpy.context.scene.alpha_trees

        row = layout.row()
        row.template_icon_view(at, "alpha_trees_previews", scale=6, scale_popup=7, show_labels=True)
        
        row = layout.row()
        row.alignment = "CENTER"
        row.label(text = at.alpha_trees_previews[:-19])

        row = layout.row()
        row.scale_y = 1.8
        row.operator("alpha_tree.import_tree", icon="VOLUME_DATA")

class ALPHATREE_PT_material_settings(bpy.types.Panel):
    """Material settings panel"""
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_region_type = "UI"
    bl_label = "Material settings"
    bl_category = "Alpha Trees"
    bl_parent_id = "ALPHATREE_PT_import_panel"

    def draw(self, context):
        layout = self.layout
        at = bpy.context.scene.alpha_trees
        object = context.active_object
        message = "   No Alpha tree selected"

        if not object:
            layout.label(text=message)
            return

        if not object.material_slots:
            layout.label(text=message)
            return
        
        if not object.material_slots[0].material:
            layout.label(text=message)
            return
        
        if not object.material_slots[0].material.use_nodes:
            layout.label(text=message)
            return
        
        if not object.material_slots[0].material.node_tree.nodes:
            layout.label(text=message)
            return

        try:
            control_node = object.material_slots[0].material.node_tree.nodes["Alpha trees control"]
        except KeyError:
            layout.label(text=message)
            return

        col = layout.column(align = True)
        imp_functions.draw_material_settings(context,col, control_node)


class ALPHATREE_PT_gen_panel(bpy.types.Panel):
    """Generator panel"""
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_region_type = "UI"
    bl_label = "Alpa trees generator"
    bl_category = "Alpha Trees"
    bl_options     = {"DEFAULT_CLOSED"}

    def draw(self, context):
        prefs = context.preferences.addons[__package__].preferences
        at = bpy.context.scene.alpha_trees
        layout = self.layout
        if prefs.show_extra_operators:
            gen_functions.dropdown_operator(
                layout,
                context,
                name="  Alpha Trees",
                operators=["alpha_tree.set_up_tree","alpha_tree.render_maps", "alpha_tree.open_folder", ],
                icons=["SORT_DESC","RESTRICT_RENDER_OFF", "FILEBROWSER", ],
                operator_booleans=[at.show_setup_tree_settings,at.show_render_settings, at.show_open_folder_settings, ],
                operator_boolean_strings=["show_setup_tree_settings", "show_render_settings", "show_open_folder_settings", ],
                operator_settings=[["z_dimension", "scale_padding", "border_padding"], [
                    "resolution", "render_filepath", "diff_render", "nor_render", "mask_render", "overwrite", "remove_extra_masks", ], ["open_leaf_folder"]],
                setttings_toggle=[[], [False, False,True, True, True, False, False], []]
            )
        
        # else:
        #     row = layout.row()
        #     row.alignment = "RIGHT"
        #     row.prop(at, "show_info", text="", icon="INFO", emboss=False)
        #     col = layout.column()
        #     col.scale_y = 0.8
        #     if at.show_info:
        #         functions.draw_info(col)

        gen_functions.operator_settings(
            layout,
            context,
            bool=at.show_exec_all_settings,
            bool_string="show_exec_all_settings",
            operator="alpha_tree.execute_all",
            settings=[["Scale to camera"], "z_dimension", "scale_padding", ["Render border"], "border_padding", ["Rendering"],
                    "resolution", "render_filepath", "diff_render", "nor_render", "mask_render", "remove_extra_masks", "overwrite", ["Open renders folder"], "open_renders_folder"],
            settings_toggle = [False,False,False,False,False,False,False,False,True,True,True,False,False,False,False,],
            icon="DRIVER",
            scale=1.75)

classes = [
    ALPHATREE_PT_import_panel,
    ALPHATREE_PT_gen_panel,
    ALPHATREE_PT_material_settings,
    AlphaTreesPrefs,
]

def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
